"""Utility functions for controlling the jj cli."""

import logging
import random
import string
import subprocess
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import Literal

logger = logging.getLogger(__name__)


@dataclass(slots=True, frozen=True)
class JJError(Exception):
    message: str
    returncode: int


def jj(
    args: list[str],
    snapshot: bool = True,
    suppress_stderr: bool = False,
    capture_stderr: bool = False,
    color: Literal["always", "never", "debug", "auto"] | None = "never",
) -> str:
    """Run a `jj` CLI command and capture its output."""
    cmd = ["jj", *args]
    if not snapshot:
        cmd.extend(["--ignore-working-copy"])
    if color is not None:
        cmd.extend(["--color", color])
    logger.debug(cmd)

    if capture_stderr:
        stderr = subprocess.STDOUT
    elif suppress_stderr:
        stderr = subprocess.PIPE  # still want to capture it for error reporting
    else:
        stderr = None  # let it go through to the terminal

    try:
        ret = subprocess.run(cmd, stderr=stderr, stdout=subprocess.PIPE, check=True)
    except subprocess.CalledProcessError as e:
        raise JJError((e.stdout or b"").decode(), e.returncode) from e
    else:
        return ret.stdout.decode().strip()


def workspace_root() -> Path:
    return Path(jj(["workspace", "root"], snapshot=False).strip())


@dataclass(slots=True, frozen=True)
class ChangeSummary:
    change_id: str
    commit_id: str
    empty: bool


def current_change() -> ChangeSummary:
    return get_changes("@")[0]


def get_changes(revset: str) -> list[ChangeSummary]:
    template = r'empty ++ "," ++ change_id.shortest(8) ++ "," ++ commit_id.shortest(12) ++ "\n"'
    output = jj(["log", "--no-graph", "-r", revset, "-T", template])
    return [
        ChangeSummary(change_id=change_id, commit_id=commit_id, empty=empty == "true")
        for line in output.splitlines()
        for empty, change_id, commit_id in [line.split(",")]
    ]


def new(ref: str | None = None):
    cmd = ["new", "--quiet"]
    if ref:
        cmd.append(ref)
    jj(cmd)


@contextmanager
def autostash():
    """Remember the working copy commit and return to it at the end of the context."""
    # Create a temporary bookmark so the current change isn't automatically abandoned
    tempbm = "jj-pre-push-keep-" + "".join(random.choices(string.ascii_letters, k=10))
    jj(["bookmark", "create", tempbm, "-r", "@", "--quiet"])
    try:
        yield
    finally:
        jj(["edit", tempbm, "--quiet"])
        jj(["bookmark", "forget", tempbm, "--quiet"])



import logging
import re
from dataclasses import dataclass
from typing import Literal

from .jj import jj

logger = logging.getLogger(__name__)

BookmarkUpdateType = Literal[
    "move_forward", "move_backward", "move_sideways", "add", "delete"
]


@dataclass(slots=True, frozen=True)
class BookmarkUpdate:
    remote: str
    bookmark: str
    update_type: BookmarkUpdateType
    old_commit: str | None = None
    new_commit: str | None = None

    def __str__(self):
        old = f" from {self.old_commit[:8]}" if self.old_commit else ""
        new = f" to {self.new_commit[:8]}" if self.new_commit else ""
        return f"{self.update_type.replace('_', ' ').capitalize()} {self.bookmark}{old}{new}"


# This is extremely brittle but it seems much better than reimplementing all the
# bookmark selection logic of `jj git push` for every possible combination of
# arguments.
# TODO: Consider parsing more rigidly?
# If the wording changes in the future, it's better to fail than to parse incorrectly!
_remote_pattern = re.compile(r"^Changes to push to (.+?):")
_bookmark_update_patterns: dict[BookmarkUpdateType, re.Pattern] = {
    "move_forward": re.compile(
        r"Move forward bookmark (?P<bookmark>\S+) from (?P<old_commit>\w+) to (?P<new_commit>\w+)"
    ),
    "move_backward": re.compile(
        r"Move backward bookmark (?P<bookmark>\S+) from (?P<old_commit>\w+) to (?P<new_commit>\w+)"
    ),
    "move_sideways": re.compile(
        r"Move sideways bookmark (?P<bookmark>\S+) from (?P<old_commit>\w+) to (?P<new_commit>\w+)"
    ),
    "add": re.compile(r"Add bookmark (?P<bookmark>\S+) to (?P<new_commit>\w+)"),
    "delete": re.compile(r"Delete bookmark (?P<bookmark>\S+) from (?P<old_commit>\w+)"),
}


def parse_git_push_dry_run(output: str) -> set[BookmarkUpdate]:
    """Extract bookmark change details from `jj git push --dry-run`."""

    updates = set()
    remote = None
    for line in output.splitlines():
        if match := _remote_pattern.search(line):
            remote = match.group(1)
        for update_type, pattern in _bookmark_update_patterns.items():
            if match := pattern.search(line):
                if remote is None:
                    raise ValueError(
                        "Unexpected line ordering in jj git push --dry-run"
                    )
                updates.add(
                    BookmarkUpdate(
                        **match.groupdict(), remote=remote, update_type=update_type
                    )
                )
    return updates


def get_remote_bookmark_updates(jj_git_push_args: list[str]) -> set[BookmarkUpdate]:
    """Given a list of CLI arguments to `jj git push`, determine the set of bookmark
    updates that would be pushed to the git remote."""
    args = ["git", "push", "--dry-run", *jj_git_push_args]
    output = jj(args, snapshot=False, capture_stderr=True)
    logger.debug(f"Output of {args}:\n{output}")
    return parse_git_push_dry_run(output)
